@@ -17,15 +17,7 @@ MODULE_DESCRIPTION("Filesystem in Userspace");
 MODULE_LICENSE("GPL");
 #endif
 
-spinlock_t fuse_lock = SPIN_LOCK_UNLOCKED;
-
-/* Must be called with the fuse lock held */
-void fuse_release_conn(struct fuse_conn *fc)
-{
-	if(fc->sb == NULL && fc->file == NULL) {
-		kfree(fc);
-	}
-}
+spinlock_t fuse_lock;
 
 int __init fuse_init(void)
 {
@@ -35,12 +27,13 @@ int __init fuse_init(void)
 	       FUSE_VERSION,
 	       FUSE_KERNEL_VERSION, FUSE_KERNEL_MINOR_VERSION);
 
+	spin_lock_init(&fuse_lock);
 	res = fuse_fs_init();
-	if(res)
+	if (res)
 		goto err;
 	
 	res = fuse_dev_init();
-	if(res)
+	if (res)
 		goto err_fs_cleanup;
 	
 	return 0;
