@@ -10,17 +10,30 @@
 #include <stdio.h>
 #include <pthread.h>
 
+/* FUSE flags: */
+
+/** Enable debuging output */
+#define FUSE_DEBUG       (1 << 1)
+
+/** If a file is removed but it's still open, don't hide the file but
+    remove it immediately */
+#define FUSE_HARD_REMOVE (1 << 2)
+
+
 typedef unsigned long fino_t;
 
 struct node {
     struct node *name_next;
     struct node *ino_next;
     fino_t ino;
+    unsigned int generation;
     fino_t parent;
     char *name;
     int mode;
     int rdev;
     int version;
+    int open_count;
+    int is_hidden;
 };
 
 struct fuse {
@@ -32,13 +45,15 @@ struct fuse {
     struct node **ino_table;
     size_t ino_table_size;
     fino_t ctr;
+    unsigned int generation;
+    unsigned int hidectr;
+    unsigned long fh_ctr;
     pthread_mutex_t lock;
     int numworker;
     int numavail;
-    struct fuse_context *(*getcontext)(struct fuse *);
-    struct fuse_context context;
-    pthread_key_t context_key;
     volatile int exited;
+    int majorver;
+    int minorver;
 };
 
 struct fuse_dirhandle {
@@ -48,7 +63,6 @@ struct fuse_dirhandle {
 };
 
 struct fuse_cmd {
-    struct fuse *f;
     char *buf;
     size_t buflen;
 };
